"""Selects overconsumption according to various criteria.

The possible criteria to select overconsumption are the following:

- :py:func:`by_individual_proportion` selects those which energy content (over
  their reference) is beyond a certain proportion of the total overconsumption
  energy. It can be used to keep only "big enough" overconsumptions.
- :py:func:`by_cumulated_proportion` selects the minimum set of overconsumption which
  are necessary to explain a certain proportion of the total overconsumption
  energy. It can be used when one wants to keep all the overconsumption which explain
  "most of the overconsumption", but can include small non-significant overshoots
  in certain cases.
- :py:func:`by_combined_proportions` is the combination of both previous approaches.
  It enables the selection of the most significant overconsumption to explain "most of
  the overconsumption".

"""

import pandas as pd


def by_individual_proportion(
    intervals_overshoot: pd.DataFrame,
    proportion_tshd: float = 0.05,
    energy_reference: float | None = None,
) -> pd.DataFrame:
    """Return overconsumption for which overshoot is above a proportion of the total.

    Parameters
    ----------
    intervals_overshoot : pd.DataFrame
        A table of overshoot overconsumption with at least 'start', 'end' and 'energy'
        columns.
    proportion_tshd : float, optional
        Proportion (in [0,1]) of the total energy of the intervals_overshoot.
        An interval is conserved by the function only if it represents at least
        (>=) this proportion of the total. Default is 5%.
    energy_reference : float or None, optional
        The total energy (in the same unit as the values in the ``"energy"``
        column, relatively to which the proportions are computed. Default is None,
        in which case the sum of the column values is used.

    Returns
    -------
    pd.DataFrame :
        A copy of the input dataframe with an additional "proportion" column, where only
        the overconsumption which represent at least the specified proportion of the
        total energy are conserved.
        The returned overconsumption are sorted by decreasing order of overshoot energy.

    """
    intervals_overshoot = intervals_overshoot.copy()
    if energy_reference is None:
        energy_reference = intervals_overshoot["energy"].sum()
    # Select overconsumption with prominent overconsumption
    intervals_overshoot["proportion"] = intervals_overshoot["energy"] / energy_reference
    intervals_overshoot = intervals_overshoot.sort_values(
        by="energy",
        ascending=False,
    )
    return intervals_overshoot[intervals_overshoot["proportion"] >= proportion_tshd]


def by_cumulated_proportion(
    intervals_overshoot: pd.DataFrame,
    proportion_tshd: float | None = 0.80,
    energy_reference: float | None = None,
) -> pd.DataFrame:
    """Return overconsumption where total energy is a proportion of the overshoot.

    Parameters
    ----------
    intervals_overshoot : pd.DataFrame
        A table of overshoot overconsumption with at least 'start', 'end' and 'energy'
        columns.
    proportion_tshd : float, optional
        Proportion (in [0,1]) of the total energy of the intervals_overshoot.
        Intervals are sorted by decreasing order of energy and conserved until
        this proportion - at least (>=) - of the total energy is reached.
        Default is 0.8, meaning that the minimum set of overconsumption which represent
        at least 80% of the total overshoot energy is conserved.
    energy_reference : float or None, optional
        The total energy (in the same unit as the values in the ``"energy"``
        column), relatively to which the proportions are computed. Default is None,
        in which case the sum of the column values is used.

    Returns
    -------
    pd.DataFrame :
        A copy of the input dataframe with an additional "cum_energy_prop" column, where
        only the overconsumption which represent together at least the specified
        proportion of the total energy are conserved.
        The returned overconsumption are sorted by decreasing order of overshoot energy.

    """
    intervals_overshoot = intervals_overshoot.sort_values(by="energy", ascending=False)
    if energy_reference is None:
        energy_reference = intervals_overshoot["energy"].sum()
    intervals_overshoot["cum_energy_prop"] = intervals_overshoot["energy"].cumsum()
    intervals_overshoot["cum_energy_prop"] /= energy_reference
    # ensure at least one value
    last_selected = (
        intervals_overshoot["cum_energy_prop"] < proportion_tshd
    ).argmin() + 1
    return intervals_overshoot.iloc[:last_selected, :]


def by_combined_proportions(
    intervals_overshoot: pd.DataFrame,
    proportion_tshd: float = 0.80,
    proportion_indiv_tshd: float = 0.05,
    energy_reference: float | None = None,
) -> pd.DataFrame:
    """Return overshoot intervals by their energy contribution based on proportions.

    Parameters
    ----------
    intervals_overshoot : pd.DataFrame
        A table of overshoot overconsumption with at least 'start', 'end' and 'energy'
        columns.
    proportion_tshd : float, optional
        Proportion (in [0,1]) of the total energy of the intervals_overshoot.
        Intervals are sorted by decreasing order of energy and conserved until
        this proportion - at least - of ``energy_reference`` is reached.
        Default is 0.8, meaning that the minimum set of intervals which represent
        at least 80% of the total overshoot energy is conserved.
    proportion_indiv_tshd : float, optional
        Proportion (in [0,1]) of the total energy of the intervals_overshoot.
        An interval is conserved by the function only if it represents at least
        this proportion of the ``energy_reference``. Default is 5%.
    energy_reference : float or None, optional
        The total energy (in the same unit as the values in the ``"energy"``
        column, relatively to which the proportions are computed. Default is None,
        in which case the sum of the column values is used.

    Returns
    -------
    pd.DataFrame :
        A copy of the input dataframe with additional "proportion" and "cum_energy_prop"
        columns, where only the overconsumption which represent together and
        individually at least the specified proportions of the total energy, are
        conserved. The returned overconsumption are sorted by decreasing order of
        overshoot energy.

    """
    if energy_reference is None:
        energy_reference = intervals_overshoot["energy"].sum()
    intervals_overshoot = intervals_overshoot.sort_values(by="energy", ascending=False)
    intervals_overshoot["cum_energy_prop"] = intervals_overshoot["energy"].cumsum()
    intervals_overshoot["cum_energy_prop"] /= energy_reference
    intervals_overshoot["proportion"] = intervals_overshoot["energy"] / energy_reference
    # ensure at least one value
    last_selected = (
        intervals_overshoot["cum_energy_prop"] < proportion_tshd
    ).argmin() + 1
    selected = intervals_overshoot.iloc[:last_selected, :]
    return selected[selected["proportion"] >= proportion_indiv_tshd]



""" """

import numpy as np
import pandas as pd
import pytest

from .. import keywords as EATK
from ..errors import (
    EATEmptyDataError,
    EATInvalidTimestepDurationError,
    EATUndefinedTimestepError,
)
from ..timeseries.extract_features.basics import (
    intervals_over,
    timestep_durations,
)

# =============================================================================
# timestep_durations
# =============================================================================


def test_evenly_spaced_durations():
    """Check the function on evenly spaced series."""
    series = pd.Series(
        np.ones(100),
        index=pd.date_range(start="2021-01-06", freq="10s", periods=100),
    )
    durs = timestep_durations(series)
    assert (durs == 10).all()
    assert (series.index == durs.index).all()
    durs = timestep_durations(series, last_step=42.0)
    assert durs.iloc[-1] == 42.0
    assert (durs.iloc[:-1] == 10).all()
    assert (series.index == durs.index).all()


def test_randomly_spaced_durations():
    """Check the function on non-evenly spaced series."""
    start = pd.Timestamp.now()
    s = pd.Timedelta("1s")
    series = pd.Series(
        np.ones(4),
        index=pd.DatetimeIndex(
            [
                start,
                start + 10 * s,
                start + 14 * s,
                start + 54 * s,
            ],
        ),
    )
    durs = timestep_durations(series)
    assert np.allclose(durs.values, np.array([10, 4, 40, 40]))
    assert (series.index == durs.index).all()


def test_durations_limit_cases():
    """Check how the functions returns with :

    - series of 0 or one element.
    - < 0s last step duration.

    """
    series = pd.Series(
        np.ones(1),
        index=pd.date_range(start="2021-01-06", freq="10s", periods=1),
    )
    with pytest.raises(EATEmptyDataError):
        timestep_durations(series.iloc[:0])
    with pytest.raises(EATUndefinedTimestepError):
        timestep_durations(series.iloc[:1], last_step=None)
    with pytest.raises(EATInvalidTimestepDurationError):
        timestep_durations(series.iloc[:1], last_step=-42.0)
    durs = timestep_durations(series.iloc[:1], last_step=42.0)
    assert durs.size == 1
    assert durs.loc["2021-01-06"] == 42.0


# =============================================================================
# ----- intervals_over
# =============================================================================


def test_intervals_over():
    """A simple test for intervals_over function."""
    time_begin = pd.Timestamp("2018-07-06 05:00:00")
    time_range = pd.date_range(
        start=time_begin,
        periods=8,
        inclusive="left",
        freq=pd.DateOffset(seconds=300),
    )
    power = pd.Series(np.array([0, 1, 1.5, 2.0, 2.0, 3.0, 1.5, 0.0]), index=time_range)
    start_f = "start"
    end_f = "end"
    up_loc, up_iloc = intervals_over(power, low_tshd=1.5, return_positions=True)
    assert up_loc.shape[0] == 1
    assert up_iloc.loc[0, start_f] == 3
    assert up_iloc.loc[0, end_f] == 6
    assert up_loc.loc[0, start_f] == pd.Timestamp("2018-07-06 05:15:00")
    assert up_loc.loc[0, end_f] == pd.Timestamp("2018-07-06 05:30:00")


def test_intervals_over_no_measures():
    """Check empty input measures return empty overconsumption."""
    power = pd.Series([])
    assert intervals_over(power, low_tshd=0.0).empty


@pytest.mark.parametrize("timestep", [1.0, 3.0, 5.0])
def test_intervals_over_no_interval(timestep):
    """Check no spurious overconsumption are found"""
    time_begin = pd.Timestamp.now()
    total_duration = 12.0 * 3600.0
    time_range = pd.date_range(
        start=time_begin,
        end=time_begin + pd.Timedelta(seconds=total_duration),
        inclusive="left",
        freq=pd.DateOffset(seconds=timestep),
    )
    power = pd.Series(np.zeros((time_range.size), dtype=np.float64), index=time_range)
    assert intervals_over(power, low_tshd=0.0).empty


@pytest.mark.parametrize("timestep", [1.0, 3.0, 5.0])
def test_intervals_over_no_interval_noisy(timestep):
    """Check no spurious heatings are found with noisy signal and threshold"""
    time_begin = pd.Timestamp.now()
    total_duration = 12.0 * 3600.0
    time_range = pd.date_range(
        start=time_begin,
        end=time_begin + pd.Timedelta(seconds=total_duration),
        inclusive="left",
        freq=pd.DateOffset(seconds=timestep),
    )
    n_steps = time_range.size
    max_noise = 50.0
    fake_data = np.abs(np.random.normal(0, 10.0, n_steps))
    fake_data = np.where(fake_data < max_noise, fake_data, max_noise)
    fake_series = pd.Series(fake_data, index=time_range)
    assert intervals_over(fake_series, low_tshd=max_noise + 1.0).empty


def test_intervals_over_2_timesteps():
    """Check on a very simple case when the periods last two timesteps"""
    time_begin = pd.Timestamp("2018-07-06 05:00:00")
    time_range = pd.date_range(
        start=time_begin,
        periods=5,
        inclusive="left",
        freq=pd.DateOffset(seconds=300),
    )
    power = pd.Series(np.array([0, 1, 1, 0, 0]) * 1000, index=time_range)
    up_loc = intervals_over(power, low_tshd=0.0)
    start = up_loc[EATK.start_f].iloc[0]
    assert up_loc.shape[0] == 1
    assert start == pd.Timestamp("2018-07-06 05:05:00")
    assert up_loc[EATK.end_f].iloc[0] == pd.Timestamp("2018-07-06 05:15:00")


def test_intervals_over_one_slot_only():
    """Check on a very simple case when the period over threshold lasts one timestep"""
    time_begin = pd.Timestamp("2018-07-06 05:00:00")
    time_range = pd.date_range(
        start=time_begin,
        periods=5,
        inclusive="left",
        freq=pd.DateOffset(seconds=300),
    )
    power = pd.Series(np.array([0, 0, 1, 0, 0]) * 1000, index=time_range)
    up_loc = intervals_over(power, low_tshd=0.0)
    start = up_loc[EATK.start_f].iloc[0]
    assert up_loc.shape[0] == 1
    assert start == pd.Timestamp("2018-07-06 05:10:00")
    assert up_loc[EATK.end_f].iloc[0] == pd.Timestamp("2018-07-06 05:15:00")


def test_intervals_over_last_step_front():
    """Check on a very simple case with period finishing on the last time-step"""
    time_begin = pd.Timestamp("2018-07-06 05:00:00")
    time_range = pd.date_range(
        start=time_begin,
        periods=5,
        inclusive="left",
        freq=pd.DateOffset(seconds=300),
    )
    power = pd.Series(np.array([0, 1, 1, 1, 0]) * 1000, index=time_range)
    up_loc = intervals_over(power, low_tshd=0.0)
    start = up_loc[EATK.start_f].iloc[0]
    assert up_loc.shape[0] == 1
    assert start == pd.Timestamp("2018-07-06 05:05:00")
    assert up_loc[EATK.end_f].iloc[0] == pd.Timestamp("2018-07-06 05:20:00")


def test_intervals_over_left_overlap():
    """Check on a very simple case with period overlapping left bound"""
    time_begin = pd.Timestamp("2018-07-06 05:00:00")
    time_range = pd.date_range(
        start=time_begin,
        periods=5,
        inclusive="left",
        freq=pd.DateOffset(seconds=300),
    )
    power = pd.Series(np.array([1, 1, 1, 0, 0]) * 1000, index=time_range)
    up_loc = intervals_over(power, low_tshd=0.0)
    start = up_loc[EATK.start_f].iloc[0]
    assert up_loc.shape[0] == 1
    assert start == pd.Timestamp("2018-07-06 05:00:00")


def test_intervals_over_right_overlap():
    """Check on a very simple case with period overlapping right bound"""
    time_begin = pd.Timestamp("2018-07-06 05:00:00")
    time_range = pd.date_range(
        start=time_begin,
        periods=5,
        inclusive="left",
        freq=pd.DateOffset(seconds=300),
    )
    power = pd.Series(np.array([0, 0, 1, 1, 1]) * 1000, index=time_range)
    up_loc = intervals_over(power, low_tshd=0.0)
    start = up_loc[EATK.start_f].iloc[0]
    assert up_loc.shape[0] == 1
    assert start == pd.Timestamp("2018-07-06 05:10:00")
    assert up_loc[EATK.end_f].iloc[0] == pd.Timestamp("2018-07-06 05:20:00")


def test_intervals_over_always_over():
    """Check on a very simple case when the value is always over the threshold"""
    time_begin = pd.Timestamp("2018-07-06 05:00:00")
    time_range = pd.date_range(
        start=time_begin,
        periods=5,
        inclusive="left",
        freq=pd.DateOffset(seconds=300),
    )
    power = pd.Series(np.array([1, 1, 1, 1, 1]) * 1000, index=time_range)
    up_loc = intervals_over(power, low_tshd=0.0)
    start = up_loc[EATK.start_f].iloc[0]
    assert up_loc.shape[0] == 1
    assert start == pd.Timestamp("2018-07-06 05:00:00")
    assert up_loc[EATK.end_f].iloc[0] == pd.Timestamp("2018-07-06 05:20:00")



"""Analyze daily-sampled thermosensitivity data."""

import logging
from collections.abc import Callable

import numpy as np
import pandas as pd

from energy_analysis_toolbox.weather.degree_days import (
    literal_computation_dd_types,
    literal_dd_types,
)

from .thermosensitivity import CategoricalThermoSensitivity


class DailyCategoricalThermoSensitivity(
    CategoricalThermoSensitivity,
):
    """Class for daily analysis of thermosensitivity data.

    Based on CategoricalThermoSensitivity, it is made to categorize the days.

    Example:
    --------
    See :py:class:`DayOfWeekCategoricalThermoSensitivity`

    """

    def __init__(
        self,
        energy_data: pd.Series,
        temperature_data: pd.Series,
        categories_func: Callable[[pd.DatetimeIndex], pd.Series],
        degree_days_type: literal_dd_types = "heating",
        degree_days_base_temperature: dict | None = None,
        degree_days_computation_method: literal_computation_dd_types = "integral",
        interseason_mean_temperature: float = 20,
        base_logger_name: str | None = None,
        min_logger_level_stdout: int | str = logging.ERROR,
    ) -> None:
        """Initialize a ``DailyCategoricalThermoSensitivity`` instance.

        Parameters
        ----------
        energy_data : pd.Series
            Time series of energy consumption data for the building.
        temperature_data : pd.Series
            Time series of outdoor temperature data.
        categories_func : Callable[[pd.DatetimeIndex], pd.Series]
            A function that takes a ``pd.DatetimeIndex`` (representing days in the
            analysis period) and returns a ``pd.Series`` that assigns a category to
            each day. This allows the analysis to group energy data by custom-defined
            categories (e.g., weekdays vs. weekends, seasons).
        degree_days_type : str, optional
            Type of degree days to compute (default is "heating").
            Options are:

            - "heating": heating degree days.
            - "cooling": cooling degree days.
            - "both": both heating and cooling degree days.
            - "auto": automatically detect the degree days type.

        degree_days_base_temperature : dict, optional
            Base temperature(s) for degree day calculations (default is an empty dict).
            Should include keys "heating" and/or "cooling".
        degree_days_computation_method : str, optional
            Method to compute degree days (default is "integral").
            Options are:

            - "integral": integral calculation.
            - "mean": mean temperature calculation.
            - "min_max": min-max temperature calculation.
            - "pro": energy-professionals calculation.

        interseason_mean_temperature : float, optional
            Mean temperature to differentiate heating and cooling periods (default 20).
        base_logger_name : str, optional
            Name of the logger. By default, it is the class name. All following
            following instances receive a unique identifier, based on the first
            one, with the pattern:

            - ``<base_logger_name>``
            - ``<base_logger_name>_1``
            - ...

        min_logger_level_stdout: str, int, optional
            Minimum logger level below which no message is transferred to stdout
            (i.e. not printed). Default is ``"ERROR"``.

        """
        frequency = "1D"
        degree_days_base_temperature = degree_days_base_temperature or {}
        start_ts = min(energy_data.index.min(), temperature_data.index.min())
        end_ts = max(energy_data.index.max(), temperature_data.index.max())
        days = pd.date_range(
            start=start_ts,
            end=end_ts,
            freq=frequency,
            inclusive="both",
        )
        categories = categories_func(days)
        super().__init__(
            energy_data=energy_data,
            temperature_data=temperature_data,
            categories=categories,
            frequency=frequency,
            degree_days_type=degree_days_type,
            degree_days_base_temperature=degree_days_base_temperature,
            degree_days_computation_method=degree_days_computation_method,
            interseason_mean_temperature=interseason_mean_temperature,
            base_logger_name=base_logger_name,
            min_logger_level_stdout=min_logger_level_stdout,
        )


class DayOfWeekCategoricalThermoSensitivity(
    DailyCategoricalThermoSensitivity,
):
    """Models independently the 7 days of the week.

    Based on :py:class:`DailyCategoricalThermoSensitivity`.
    """

    def __init__(
        self,
        energy_data: pd.Series,
        temperature_data: pd.Series,
        degree_days_type: literal_dd_types = "heating",
        degree_days_base_temperature: dict | None = None,
        degree_days_computation_method: literal_computation_dd_types = "integral",
        interseason_mean_temperature: float = 20,
        base_logger_name: str | None = None,
        min_logger_level_stdout: int | str = logging.ERROR,
    ) -> None:
        """Initialize a ``DayOfWeekCategoricalThermoSensitivity`` instance.

        Parameters
        ----------
        energy_data : pd.Series
            Time series of energy consumption data for the building.
        temperature_data : pd.Series
            Time series of outdoor temperature data.
        degree_days_type : str, optional
            Type of degree days to compute (default is "heating").
            Options are:

            - "heating": heating degree days.
            - "cooling": cooling degree days.
            - "both": both heating and cooling degree days.
            - "auto": automatically detect the degree days type.

        degree_days_base_temperature : dict, optional
            Base temperature(s) for degree day calculations (default is an empty dict).
            Should include keys "heating" and/or "cooling".
        degree_days_computation_method : str, optional
            Method to compute degree days (default is "integral").
            Options are:

            - "integral": integral calculation.
            - "mean": mean temperature calculation.
            - "min_max": min-max temperature calculation.
            - "pro": energy-professionals calculation.

        interseason_mean_temperature : float, optional
            Mean temperature to differentiate heating and cooling periods (default 20).
        base_logger_name : str, optional
            Name of the logger. By default, it is the class name. All following
            following instances receive a unique identifier, based on the first
            one, with the pattern:

            - ``<base_logger_name>``
            - ``<base_logger_name>_1``
            - ...

        min_logger_level_stdout: str, int, optional
            Minimum logger level below which no message is transferred to stdout
            (i.e. not printed). Default is ``"ERROR"``.

        """
        degree_days_base_temperature = degree_days_base_temperature or {}

        def day_of_week_categoriser(
            index: pd.DatetimeIndex,
        ) -> pd.Series:
            """Return a series of categories based on the day of the week of the index.

            Parameters
            ----------
            index : pd.DatetimeIndex
                A datetime index representing the dates to categorize.

            Returns
            -------
            pd.Series
                A pandas Series where each entry is the name of the day of the week
                corresponding to the respective index value.

            Examples
            --------
            >>> index = pd.date_range(start="2023-01-01", periods=7, freq="D")
            >>> day_of_week_categoriser(index)
            2023-01-01       Sunday
            2023-01-02       Monday
            2023-01-03      Tuesday
            2023-01-04    Wednesday
            2023-01-05     Thursday
            2023-01-06       Friday
            2023-01-07     Saturday
            Freq: D, dtype: object

            """
            return pd.Series(index=index, data=index.day_name())

        super().__init__(
            energy_data=energy_data,
            temperature_data=temperature_data,
            categories_func=day_of_week_categoriser,
            degree_days_type=degree_days_type,
            degree_days_base_temperature=degree_days_base_temperature,
            degree_days_computation_method=degree_days_computation_method,
            interseason_mean_temperature=interseason_mean_temperature,
            base_logger_name=base_logger_name,
            min_logger_level_stdout=min_logger_level_stdout,
        )


class AutoCategoricalThermoSensitivity(
    DayOfWeekCategoricalThermoSensitivity,
):
    """Automatically categorizes thermosensitivity data based on predefined criteria.

    Based on :py:class:`DayOfWeekCategoricalThermoSensitivity`.
    """

    @property
    def categories(
        self,
    ) -> pd.Series:
        """The categories of the periods."""
        return self._categories

    @categories.setter
    def categories(
        self,
        value: pd.Series,
    ) -> None:
        """Set the categories and reset associated cached data.

        This method assigns new category labels to the ``categories`` attribute and
        resets the cached data that depends on these categories, ensuring that all
        computations are up to date with the new categorization.

        Parameters
        ----------
        value : pd.Series
            A pandas Series representing the new categories to assign.

        .. note::
            Setting new categories automatically resets the following internal
            attributes:

            - ``resampled_energy_temperature_category``: Cached energy-temperature data
            that is resampled by category.
            - ``resampled_categories``: Cached resampled category values.
            - ``_aggregated_data``: Cached aggregated data, if any.

            These attributes are recalculated upon the next request, ensuring
            consistency with the newly assigned categories.

        """
        self._categories = value
        self.__dict__.pop("resampled_energy_temperature_category", None)
        self.__dict__.pop("resampled_categories", None)
        self._aggregated_data = None

    def new_categories(
        self,
        significant_level: float = 0.1,
    ) -> dict:
        """Return new category mappings based on interaction term significance.

        This method identifies significant differences between the thermosensitivity
        of each category and provides new mappings that merge similar categories.
        Categories are grouped based on the significance of interaction terms
        in the model, which helps reduce complexity while retaining meaningful
        distinctions.

        Parameters
        ----------
        significant_level : float, optional
            The significance level for the Wald test (a p-value below this level
            is considered significant). Must be between 0 and 1. The higher the
            value, the more categories will be kept separate. Lower values will
            merge categories that are not significantly different.

        Returns
        -------
        dict
            A dictionary mapping old categories to new merged categories.
            The new labels are concatenated with a "-" separator to indicate
            merged groups.

        Notes
        -----
        - The new categories are based on the result of multiple Wald tests conducted
          between interaction terms for each category.
        - The returned dictionary allows for updating the category labels to reflect
          merged groupings that exhibit similar behavior.


        Example
        -------
        >>> auto = AutoCategoricalThermoSensitivity(...)
        >>> auto.fit()
        >>> auto.new_categories(significant_level=0.1)
        {'Monday': 'Monday-Wednesday-Sunday',
         'Tuesday': 'Tuesday',
         'Wednesday': 'Monday-Wednesday-Sunday',
         'Thursday': 'Thursday',
         'Friday': 'Friday',
         'Saturday': 'Monday-Wednesday-Sunday',
         'Sunday': 'Sunday'
        }

        """
        categories_sorted = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]
        categories = self.resampled_categories.unique()
        predictors = [*self.predictors, "Intercept"]
        new_categories_mapping = {str(term): [str(term)] for term in categories}
        for i, cat_term1 in enumerate(categories):
            for _, cat_term2 in enumerate(categories[i + 1 :]):
                is_same_group = True
                for pred in predictors:
                    interaction_term1 = pred + ":" + cat_term1
                    interaction_term2 = pred + ":" + cat_term2
                    contrast_matrix = np.zeros((1, len(self.model.params)))
                    contrast_matrix[
                        0,
                        self.model.params.index.get_loc(interaction_term1),
                    ] = 1
                    contrast_matrix[
                        0,
                        self.model.params.index.get_loc(interaction_term2),
                    ] = -1
                    wald_test = self.model.wald_test(contrast_matrix, scalar=True)
                    if wald_test.pvalue < significant_level:
                        is_same_group &= False
                if is_same_group:
                    new_categories_mapping[cat_term1].append(cat_term2)
                    new_categories_mapping[cat_term2] = new_categories_mapping[
                        cat_term1
                    ]
        reduced_mapping = {
            k: sorted(set(v), key=lambda d: categories_sorted.index(d))
            for k, v in new_categories_mapping.items()
        }
        return {k: "-".join(v) for k, v in reduced_mapping.items()}

    def merge_and_fit(
        self,
        significant_level: float = 0.1,
    ) -> None:
        """Merge similar categories and fit the model with updated categories.

        This method merges categories that exhibit similar thermosensitivity based
        on their interaction term significance and then refits the model using
        the updated categorization. This helps reduce model complexity by grouping
        categories with similar behavior.

        Parameters
        ----------
        significant_level : float, optional
            The significance level for the Wald test (a p-value below this level
            is considered significant). Must be between 0 and 1. The higher the value,
            the more categories will be kept separate. Lower values will merge
            categories that are not significantly different.

        Returns
        -------
        None

        Notes
        -----
        - The method first calculates new categories using the ``new_categories``
          method, then assigns these new categories and fits the model to reflect the
          updated categorization.
        - This process is especially useful for reducing overfitting by combining
          similar days (e.g., merging weekdays that do not show significant
          thermosensitivity differences).

        """
        new_cats_maps = self.new_categories(significant_level=significant_level)
        self.categories = self.categories.map(new_cats_maps)
        self.fit()



""" """

import pandas as pd
import scipy.constants as SK

from energy_analysis_toolbox.timeseries.profiles.preprocessing.history_filters.weekdays import (
    same_day_only,
    weekdays_only,
    weekends_only,
)

from .fake_data import sinusoid_history_df


def test_same_day():
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday, n_days=21, freq="1D", period_variation=7 * SK.day,
    )
    # on df
    filtered = same_day_only(history, date=monday + pd.Timedelta("1D"))
    pd.testing.assert_frame_equal(filtered, history.iloc[[1, 8, 15], :])
    # on series
    filtered_s = same_day_only(history.iloc[:, 0])
    pd.testing.assert_series_equal(filtered_s, history.iloc[[0, 7, 14], 0])


def test_weekends():
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday, n_days=21, freq="1D", period_variation=7 * SK.day,
    )
    # on df
    filtered = weekends_only(history)
    pd.testing.assert_frame_equal(filtered, history.iloc[[5, 6, 12, 13, 19, 20], :])
    # on series
    filtered_s = weekends_only(history.iloc[:, 0])
    pd.testing.assert_series_equal(filtered_s, history.iloc[[5, 6, 12, 13, 19, 20], 0])


def test_weekdays():
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday, n_days=7, freq="1D", period_variation=7 * SK.day,
    )
    # on df
    filtered = weekdays_only(history)
    pd.testing.assert_frame_equal(filtered, history.iloc[0:5, :])
    # on series
    filtered_s = weekdays_only(history.iloc[:, 0])
    pd.testing.assert_series_equal(filtered_s, history.iloc[0:5, 0])


def test_weekdays_weekends_orthogonal():
    """Test that filtering weekends then weekdays only returns an empty history."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday, n_days=7, freq="1D", period_variation=7 * SK.day,
    )
    assert weekdays_only(weekends_only(history)).empty


def test_pipelining():
    """Check that the fileters can be applied through DatafRame.pipeline method."""
    monday = pd.Timestamp("2021-12-27")
    history = sinusoid_history_df(
        start=monday, n_days=14, freq="1D", period_variation=7 * SK.day,
    )
    filtered = history.pipe(same_day_only).pipe(weekdays_only)
    pd.testing.assert_frame_equal(filtered, history.iloc[[0, 7], :], check_freq=False)
    filtered = history.pipe(same_day_only).pipe(weekends_only)
    filtered.empty



"""Check the conformity of the utility functions used in draw detection
"""

import numpy as np
import pandas as pd
import pytest

from ..timeseries.math.derivatives import (
    time_derivative_fwd,
    time_derivative_second,
)
from ..timeseries.resample.index_transformation import tz_convert_or_localize
from ..timeseries.resample.interpolate import (
    piecewise_affine,
    piecewise_constant,
)

# =============================================================================
# Derivatives
# =============================================================================


@pytest.fixture
def constant_timeseries():
    """Create a timeseries of constant values"""
    nw = pd.Timestamp.now()
    inds = pd.date_range(nw, nw + pd.Timedelta(seconds=10), freq="2s")
    series = pd.Series(np.ones(inds.size), index=inds)
    return series


@pytest.fixture
def affine_timeseries():
    """Create a timeseries of affine increasing values"""
    nw = pd.Timestamp.now()
    inds = pd.date_range(nw, nw + pd.Timedelta(seconds=10), freq="2s")
    series = pd.Series(np.linspace(1, 11, inds.size), index=inds)
    return series


def test_derivative_constant(constant_timeseries):
    """Check that the derivative computed for a constant function is zero"""
    d1t = time_derivative_fwd(constant_timeseries)
    assert (d1t == 0.0).all()
    assert d1t.size == constant_timeseries.size


def test_derivative_affine(affine_timeseries):
    """Check the derivative computation for an affine function"""
    d1t = time_derivative_fwd(affine_timeseries)
    assert np.allclose(d1t.values, np.ones(affine_timeseries.size))


def test_second_derivative_constant(constant_timeseries):
    """Check that the second order derivative computed for a constant function is zero"""
    d2t = time_derivative_second(constant_timeseries)
    assert (d2t == 0.0).all()


def test_second_derivative_affine(affine_timeseries):
    """Check that the second order derivative computed for an affine function is zero"""
    d2t = time_derivative_second(affine_timeseries)
    assert (d2t == 0.0).all()


# =============================================================================
# ----- BC interpolations ------
# =============================================================================


@pytest.mark.parametrize(
    "source_tz, targets_tz",
    [
        (None, None),
        ("Europe/Paris", None),
        (None, "Europe/Paris"),
        ("Europe/Paris", "Europe/Paris"),
        ("Europe/Paris", "UTC"),
        ("UTC", "Europe/Paris"),
    ],
)
@pytest.mark.parametrize(
    "targets, expected",
    [  # samples
        (np.array([0.0, 5.0, 7.0, 10.0]), np.array([0.0, 1.0, 2.0, -1.0])),
        # between samples
        (np.array([1.0, 6.0, 9.0]), np.array([0.2, 1.5, 0.0])),
        # outside samples
        (np.array([-1, 12.0]), np.array([0.0, -1.0])),
        # mix
        (np.array([-1, 6.0, 7.0, 12.0]), np.array([0.0, 1.5, 2.0, -1.0])),
        # No values
        (np.array([]), np.array([])),
    ],
)
def test_piecewise_affine(targets, expected, source_tz, targets_tz):
    """Check the method to interpolate BC timeseries : piecewise affine

    .. note::

        As the source and target instants are generated as offsets from the
        start time, the expected values are not affected by the timezone.


    """
    start_time = pd.Timestamp.now(tz=source_tz)
    sample_dts = [0.0, 5.0, 7.0, 10.0]
    sample_values = [0.0, 1.0, 2.0, -1.0]
    instants = pd.DatetimeIndex(
        [start_time + pd.Timedelta(seconds=dt) for dt in sample_dts],
    )
    bc = pd.Series(sample_values, instants)
    target_instants = pd.DatetimeIndex(
        [start_time + pd.Timedelta(seconds=dt) for dt in targets],
    )
    tz_convert_or_localize(target_instants, targets_tz)
    affine_interp = piecewise_affine(bc, target_instants)
    assert np.allclose(affine_interp.to_numpy(), expected)


@pytest.mark.parametrize(
    "source_tz, targets_tz",
    [
        (None, None),
        ("Europe/Paris", None),
        (None, "Europe/Paris"),
        ("Europe/Paris", "Europe/Paris"),
        ("Europe/Paris", "UTC"),
        ("UTC", "Europe/Paris"),
    ],
)
@pytest.mark.parametrize(
    "targets, expected",
    [  # samples
        (np.array([0.0, 5.0, 7.0, 10.0]), np.array([0.0, 1.0, 2.0, -1.0])),
        # between samples
        (np.array([1.0, 6.0, 9.0]), np.array([0.0, 1.0, 2.0])),
        # outside samples
        (np.array([-1, 12.0]), np.array([0.0, -1.0])),
        # mix
        (np.array([-1, 6.0, 7.0, 12.0]), np.array([0.0, 1.0, 2.0, -1.0])),
        # No values
        (np.array([]), np.array([])),
    ],
)
def test_piecewise_constant(targets, expected, source_tz, targets_tz):
    """Check the method to interpolate BC timeseries : piecewise constant.

    .. note::

        As the source and target instants are generated as offsets from the
        start time, the expected values are not affected by the timezone.

    """
    start_time = pd.Timestamp.now(tz=source_tz)
    sample_dts = [0.0, 5.0, 7.0, 10.0]
    sample_values = [0.0, 1.0, 2.0, -1.0]
    instants = pd.DatetimeIndex(
        [start_time + pd.Timedelta(seconds=dt) for dt in sample_dts],
    )
    bc = pd.Series(sample_values, instants)
    target_instants = pd.DatetimeIndex(
        [start_time + pd.Timedelta(seconds=dt) for dt in targets],
    )
    tz_convert_or_localize(target_instants, targets_tz)
    stair_interp = piecewise_constant(bc, target_instants)
    assert np.allclose(stair_interp.to_numpy(), expected)
    pd.testing.assert_index_equal(stair_interp.index, target_instants)


def test_piecewise_constant_padding():
    """Check the left-padding"""
    start_time = pd.Timestamp.now()
    sample_dts = [0.0, 5.0, 7.0, 10.0]
    sample_values = [0.0, 1.0, 2.0, -1.0]
    instants = pd.DatetimeIndex(
        [start_time + pd.Timedelta(seconds=dt) for dt in sample_dts],
    )
    bc = pd.Series(sample_values, instants)
    target_instants = pd.DatetimeIndex(
        [start_time + pd.Timedelta(seconds=dt) for dt in np.array([-2, -1, 12.0])],
    )
    stair_interp = piecewise_constant(bc, target_instants)
    assert np.allclose(stair_interp.to_numpy(), np.array([0.0, 0.0, -1.0]))
    stair_interp = piecewise_constant(bc, target_instants, left_pad=42.0)
    assert np.allclose(stair_interp.to_numpy(), np.array([42.0, 42.0, -1.0]))



"""Test the degree days module."""

import numpy as np
import pandas as pd

from energy_analysis_toolbox.weather.degree_days import (
    dd_calc_method,
    dd_compute,
    dd_integral,
    dd_mean,
    dd_min_max,
    dd_pro,
)


def test_dd_mean_basic_features():
    """Verifies that, when computing degree-days, the name property is set and that we have 7 days."""
    index = pd.date_range(
        start="2023-01-21",
        periods=24 * 60 * 7,
        freq="1min",
        tz="Europe/Paris",
    )
    data = [0] * len(index)
    temperature = pd.Series(
        data=data,
        index=index,
    )
    dd = dd_mean(temperature, reference=15, dd_type="heating")
    assert dd.name == "heating_degree_days"
    assert len(dd) == 7
    assert dd.dtype == "float64"


def test_dd_mean_missing_data():
    """Verifies that, even a data is set to ``NaN``, the degree-days goes well."""
    index = pd.date_range(
        start="2023-01-21",
        periods=12 * 60,
        freq="1min",
        tz="Europe/Paris",
    )
    data = [0] * len(index)
    temperature = pd.Series(
        data=data,
        index=index,
    )
    temperature.iloc[0] = None
    dd = dd_mean(temperature, reference=15, dd_type="heating")
    expected_dd = pd.Series(
        [15.0],
        name="heating_degree_days",
        index=pd.date_range(
            start="2023-01-21",
            periods=1,
            freq="1D",
            tz="Europe/Paris",
        ),
    )
    pd.testing.assert_series_equal(dd, expected_dd)


def test_dd_min_max():
    """Verifies that, when computing degree-days, the name property is set and that we have 7 days."""
    index = pd.date_range(
        start="2023-01-21",
        periods=24 * 60 * 7,
        freq="1min",
        tz="Europe/Paris",
    )
    data = [0] * len(index)
    temperature = pd.Series(
        data=data,
        index=index,
    )
    dd = dd_min_max(temperature, reference=15, dd_type="heating")
    assert dd.name == "heating_degree_days"
    assert len(dd) == 7
    assert dd.dtype == "float64"


def test_dd_integral():
    """Verifies that, when computing degree-days, the name property is set and that we have 7 days."""
    index = pd.date_range(
        start="2023-01-21",
        periods=24 * 60 * 7,
        freq="1min",
        tz="Europe/Paris",
    )
    data = [0] * len(index)
    temperature = pd.Series(
        data=data,
        index=index,
    )
    dd = dd_integral(temperature, reference=15, dd_type="heating")
    assert dd.name == "heating_degree_days"
    assert len(dd) == 7
    assert dd.dtype == "float64"


def test_dd_compute():
    """Verifies that, when computing degree-days, the name property is set and that we have 7 days."""
    index = pd.date_range(
        start="2023-01-21",
        periods=24 * 60 * 7,
        freq="1min",
        tz="Europe/Paris",
    )
    data = [0] * len(index)
    temperature = pd.Series(
        data=data,
        index=index,
    )
    dd = dd_compute(temperature, reference=15, method="integral", dd_type="heating")
    assert dd.name == "heating_degree_days"
    assert len(dd) == 7
    assert dd.dtype == "float64"


def generate_days_sin_data(n_periods=2, frequency="1h"):
    """Generate one week of sinusoidal data."""
    start = pd.Timestamp("2023-01-21 00:00:00", tz="Europe/Paris")
    end = start + pd.Timedelta(days=n_periods, seconds=-1)
    index = pd.date_range(
        start=start,
        end=end,
        freq=frequency,
    )
    number_points = len(index)
    data = np.sin(np.linspace(0, 2 * np.pi * n_periods, number_points))
    temperature = pd.Series(
        data=data,
        index=index,
    )
    return temperature


def reference_sin_dd(mean, spread, reference=15):
    """Compute the degree days with the integral method
    for a sin function.

    .. note::

        This function uses the formula:

        .. math::

            \\int_{t_0}^{t_1} \\left( sin(t) - sin(t_0) \\right() dt = cos(t_0) - cos(t_1) + (t_1 - t_0) sin(t_0)

        The ``mean`` and ``spread`` parameters are used to scale the values.
    """
    if mean - spread > reference:
        return 0
    if mean + spread <= reference:
        return reference - mean
    ref_norm = (reference - mean) / spread
    # first time when temperature is below reference
    t0 = np.arcsin(-ref_norm)
    # second time when temperature is above reference
    t1 = np.pi - t0
    expected_dd = np.cos(t0) - np.cos(t1) + (t1 - t0) * ref_norm
    return expected_dd * spread / (2 * np.pi)


def test_computation_realistic_data_integral(spread=10, mean=15):
    """Test the value of the computation of the degree_days using integral
    for a Sin temperature evolution.

    """
    # using high sampling to reduce the error on the integral
    # as we use the rectangle integral method
    temperature = generate_days_sin_data(frequency="1min") * spread + mean
    ref_temperatures = np.linspace(5, 30, 20)
    for ref in ref_temperatures:
        # using mean as there are multiple days in the data
        computed_dd = dd_compute(
            temperature,
            reference=ref,
            method="integral",
            dd_type="heating",
        ).mean()
        # putting the reference computation in the assert to better debug the test
        assert np.isclose(
            computed_dd,
            reference_sin_dd(mean, spread, reference=ref),
            rtol=1e-2,
        )


def test_computation_realistic_data_pro(spread=10, mean=15):
    """Test the value of the computation of the degree_days using the pro method
    for a Sin temperature evolution.

    """
    temperature = generate_days_sin_data(frequency="60min") * spread + mean
    ref_temperatures = np.linspace(5, 30, 20)
    for ref in ref_temperatures:
        # using mean as there are multiple days in the data
        computed_dd = dd_compute(
            temperature,
            reference=ref,
            method="pro",
            dd_type="heating",
        ).mean()
        # putting the reference computation in the assert to better debug the test
        assert np.isclose(
            computed_dd,
            reference_sin_dd(mean, spread, reference=ref),
            rtol=0.2,
        )  # the pro method is less accurate


def reference_sin_dd_not_integral(mean, spread, reference):
    """Return the expected NOT INTEGRAL DD for a sin function temperature .

    This works for both min-max and mean methods, as both return the same value
    when the temperature is symmetric around the mean, as in the sin function.
    """
    if mean > reference:
        return 0
    if mean <= reference:
        return reference - mean


def test_computation_realistic_data_not_integral(spread=10, mean=15):
    """Test the value of the computation of the degree_days using NOT integral
    for a Sin temperature evolution.

    """
    # using high sampling to reduce the error on the integral
    # as we use the rectangle integral method
    temperature = generate_days_sin_data(frequency="1h") * spread + mean
    ref_temperatures = np.linspace(5, 30, 20)
    for ref in ref_temperatures:
        # using mean as there are multiple days in the data
        computed_dd = dd_compute(
            temperature, reference=ref, method="min_max", dd_type="heating"
        ).mean()
        # putting the reference computation in the assert to better debug the test
        assert np.isclose(
            computed_dd,
            reference_sin_dd_not_integral(mean, spread, reference=ref),
            rtol=1e-2,
        )
    for ref in ref_temperatures:
        # using mean as there are multiple days in the data
        computed_dd = dd_compute(
            temperature, reference=ref, method="mean", dd_type="heating"
        ).mean()
        # putting the reference computation in the assert to better debug the test
        assert np.isclose(
            computed_dd,
            reference_sin_dd_not_integral(mean, spread, reference=ref),
            rtol=1e-2,
        )


def test_dd_calc_method():
    """Test the method to compute the degree days."""
    assert dd_calc_method(dd_mean) == "mean"
    assert dd_calc_method(dd_min_max) == "min_max"
    assert dd_calc_method(dd_integral) == "integral"
    assert dd_calc_method(dd_pro) == "pro"
    assert dd_calc_method(dd_compute) == "unknown"
    assert dd_calc_method(np.min) == "unknown"



import pandas as pd
import pytest

from .. import pandas  # noqa:F401

# Test EATAccessorSeries


def test_series_to_energy():
    # Test the to_energy method of EATAccessorSeries
    from energy_analysis_toolbox.power import to_energy

    series = pd.Series(
        [1, 2, 3],
        index=pd.date_range("2022-01-01", periods=3, freq="D"),
    )
    ct_series = series.eat.to_energy()
    assert isinstance(ct_series, pd.Series)
    pd.testing.assert_series_equal(ct_series, to_energy(series))


def test_series_to_power():
    # Test the to_power method of EATAccessorSeries
    series = pd.Series(
        [24, 48, 72],
        index=pd.date_range("2022-01-01", periods=3, freq="D"),
    )
    with pytest.raises(NotImplementedError):
        series.eat.to_power()


def test_series_power_to_freq():
    # Test the power_to_freq method of EATAccessorSeries
    from energy_analysis_toolbox.power import to_freq

    series = pd.Series(
        [1, 2, 3],
        index=pd.date_range("2022-01-01", periods=3, freq="D"),
    )
    ct_series = series.eat.power_to_freq("2D", last_step_duration=3600)
    assert isinstance(ct_series, pd.Series)
    pd.testing.assert_series_equal(
        ct_series,
        to_freq(series, "2D", last_step_duration=3600),
    )


def test_series_energy_to_freq():
    # Test the energy_to_freq method of EATAccessorSeries
    from energy_analysis_toolbox.energy import to_freq

    series = pd.Series(
        [1, 2, 3],
        index=pd.date_range("2022-01-01", periods=3, freq="D"),
    )
    ct_series = series.eat.energy_to_freq("2D", last_step_duration=3600)
    assert isinstance(ct_series, pd.Series)
    pd.testing.assert_series_equal(
        ct_series,
        to_freq(series, "2D", last_step_duration=3600),
    )


def test_series_intervals_over():
    # Test the intervals_over method of EATAccessorSeries
    from energy_analysis_toolbox.timeseries.extract_features import (
        intervals_over,
    )

    series = pd.Series(
        [1, 2, 3],
        index=pd.date_range("2022-01-01", periods=3, freq="D"),
    )
    ct_series = series.eat.intervals_over(2)
    assert isinstance(ct_series, pd.DataFrame)
    pd.testing.assert_frame_equal(ct_series, intervals_over(series, 2))


def test_series_timestep_durations():
    # Test the timestep_durations method of EATAccessorSeries
    from energy_analysis_toolbox.timeseries.extract_features import (
        timestep_durations,
    )

    series = pd.Series(
        [1, 2, 3],
        index=pd.date_range("2022-01-01", periods=3, freq="D"),
    )
    ct_series = series.eat.timestep_durations()
    assert isinstance(ct_series, pd.Series)
    pd.testing.assert_series_equal(ct_series, timestep_durations(series))


def test_series_fill_missing_entries():
    # Test the fill_missing_entries method of EATAccessorSeries
    from energy_analysis_toolbox.timeseries.resample import fill_data_holes

    series = pd.Series(
        [1, 2, 3, 4, 5],
        index=pd.date_range("2022-01-01 01:00:00", periods=5, freq="1h"),
    )
    series = series.drop(pd.Timestamp("2022-01-01 03:00:00"))
    ct_series = series.eat.fill_data_holes()
    assert isinstance(ct_series, pd.Series)
    pd.testing.assert_series_equal(ct_series, fill_data_holes(series))


# Test EATAccessorFrame
def test_frame():
    # Test the to_energy method of EATAccessorFrame
    df = pd.DataFrame(
        {"value": [1, 2, 3], "duration": [1000, 2000, 3000]},
        index=pd.date_range("2022-01-01", periods=3, freq="D"),
    )
    with pytest.raises(NotImplementedError):
        df.eat



"""Testing module for time-localized profiles"""

import numpy as np
import pandas as pd

from energy_analysis_toolbox.timeseries.profiles.localization import (
    LocalizedMeanProfile,
    LocalizedRollingProfile,
    LocalizedRollingQuantileProfile,
    MeanProfile,
    RollingProfile,
    RollingQuantileProfile,
)


def test_localized_mean_profile():
    """Test that the Localize Mixin provide the same results as the
    non-localized profile when no DST happens between the first day of history
    and the target date:

    1. without timezone
    2. with timezone


    """
    n_days = 3
    freq_minutes = 6
    start = pd.Timestamp("2022-09-23 00:00")
    end = start + pd.DateOffset(days=n_days, milliseconds=-1)
    index = pd.DatetimeIndex(
        data=pd.date_range(start=start, end=end, freq=str(freq_minutes) + "min"),
    )
    data = np.random.randn(len(index))
    history = pd.DataFrame(
        data={"value": data},
        index=index,
    )
    time = pd.Timestamp("2022-10-23 00:00")
    expected = MeanProfile().compute(history=history, time=time)
    result = LocalizedMeanProfile().compute(history=history, time=time)
    # 1.
    pd.testing.assert_frame_equal(expected, result)
    history.index = index.tz_localize("Europe/Paris")
    result = LocalizedMeanProfile().compute(history=history, time=time)
    expected.index = expected.index.tz_localize("Europe/Paris")
    # 2.
    pd.testing.assert_frame_equal(expected, result)


def test_localized_rolling_profile():
    """Test that the Localize Mixin provide the same results as the
    Non-localized Profile:
    1. without timezone
    2. with timezone
    """
    n_days = 3
    freq_minutes = 6
    start = pd.Timestamp("2022-09-23 00:00")
    window = "30min"
    end = start + pd.DateOffset(days=n_days, milliseconds=-1)
    index = pd.DatetimeIndex(
        data=pd.date_range(start=start, end=end, freq=f"{freq_minutes}min"),
    )
    data = np.random.randn(len(index))
    history = pd.DataFrame(
        data={"value": data},
        index=index,
    )
    time = pd.Timestamp("2022-10-23 00:00")
    agg = np.std
    expected = RollingProfile(window, agg).compute(history=history, time=time)
    result = LocalizedRollingProfile(window, agg).compute(history=history, time=time)
    # 1.
    pd.testing.assert_frame_equal(expected, result)
    history.index = index.tz_localize("Europe/Paris")
    result = LocalizedRollingProfile(window, agg).compute(history=history, time=time)
    expected.index = expected.index.tz_localize("Europe/Paris")
    # 2.
    pd.testing.assert_frame_equal(expected, result)


def test_localized_rolling_quantile_profile():
    """Test that the Localize Mixin provide the same results as the
    Non-localized Profile:
    1. without timezone
    2. with timezone
    """
    n_days = 3
    freq_minutes = 6
    start = pd.Timestamp("2022-09-23 00:00")
    window = "30min"
    q = 0.8
    end = start + pd.DateOffset(days=n_days, milliseconds=-1)
    index = pd.DatetimeIndex(
        data=pd.date_range(start=start, end=end, freq=str(freq_minutes) + "min"),
    )
    data = np.random.randn(len(index))
    history = pd.DataFrame(
        data={"value": data},
        index=index,
    )
    time = pd.Timestamp("2022-10-23 00:00")
    expected = RollingQuantileProfile(window, q).compute(history=history, time=time)
    result = LocalizedRollingQuantileProfile(window, q).compute(
        history=history, time=time,
    )
    # 1.
    pd.testing.assert_frame_equal(expected, result)
    history.index = index.tz_localize("Europe/Paris")
    result = LocalizedRollingQuantileProfile(window, q).compute(
        history=history, time=time,
    )
    expected.index = expected.index.tz_localize("Europe/Paris")
    # 2.
    pd.testing.assert_frame_equal(expected, result)



r"""Provides functions for calculating degree days.

The provided functions assumes that:

- the sampling is not regular (missing data possible)
- the temperature is given in Celsius
- the frequency of the data is around hourly (e.g. 1-3 hours), as the computations
  assume multiple temperature values during each day to compute the ``min``, ``max``,
  ``integral``, and so on during each day.

Notation
--------
- :math:`DD` : Degree-days for a given day (heating or cooling)
- :math:`T_{min}` : Minimum temperature for a given day
- :math:`T_{max}` : Maximum temperature for a given day
- :math:`T_{ref}` : Reference temperature for degree-days computation
- :math:`\\Delta t_i` : Duration of the i-th timestep
- :math:`T_{mean}` : Mean temperature for a given day computed as
  :math:`\\sum_{i=1}^{N} T_i \\cdot \\Delta t_i / \\sum_{i=1}^{N} \\Delta t_i`
- :math:`N` : Number of timesteps in a day
- :math:`T_i` : Temperature at the i-th timestep
- :math:`(x)^+` : Positive part of x, i.e. :math:`max(x, 0)`
- Type of degree-days:

  - Heating : Degree-days for heating season, means that the reference temperature is
    higher than the outside temperature
  - Cooling : Degree-days for cooling season, means that the reference temperature is
    lower than the outside temperature

Examples
--------
>>> temp = pd.Series(np.random.randn(100), index=pd.date_range("2020-01-01",
... periods=100, freq='h'))
>>> dd_compute(temp, 17, method="integral").head()
2020-01-01    17.147638
2020-01-02    17.262956
2020-01-03    16.903173
2020-01-04    17.132581
2020-01-05    16.918075
Freq: D, Name: heating_degree_days, dtype: float64

>>> dd_compute(temp, 17, method="integral", type="cooling").head()
2020-01-01    0.0
2020-01-02    0.0
2020-01-03    0.0
2020-01-04    0.0
2020-01-05    0.0
Freq: D, Name: cooling_degree_days, dtype: float64

"""

from collections.abc import Callable
from typing import Literal, cast

import pandas as pd

import energy_analysis_toolbox as eat
from energy_analysis_toolbox.errors.degree_days import (
    EATInvalidDegreeDaysError,
    EATInvalidDegreeDaysMethodError,
)

dd_types = [
    "heating",
    "cooling",
    "both",
    "auto",
]
computation_dd_types = [
    "min_max",
    "mean",
    "integral",
    "pro",
]
literal_dd_types = Literal[
    "heating",
    "cooling",
    "both",
    "auto",
]
literal_valid_dd_types = Literal[
    "heating",
    "cooling",
]
literal_computation_dd_types = Literal[
    "min_max",
    "mean",
    "integral",
    "pro",
]

DegreeDaysFunction = Callable[
    [
        pd.Series,
        float,
        literal_valid_dd_types,
        float,
    ],  # Arguments: temperature, reference, clip_tshd, dd_type
    pd.Series,  # Return type
]


def dd_min_max(
    temperature: pd.Series,
    reference: float,
    dd_type: literal_valid_dd_types,
    clip_tshd: float = 0,
) -> pd.Series:
    r"""Return daily degree-days with min-max method.

    .. math::
        DD = (\\frac{T_{min} + T_{max}}{2} - T_{ref})^+

    Parameters
    ----------
    temperature : pd.Series
        The timeseries of temperature measures from which DD data has to be inferred.
    reference : float
        The reference temperature for  degree-days computation.
    dd_type : {'heating', 'cooling'}
        The type of degree-days to compute.
    clip_tshd : float or None, optional
        A threshold under which the computed degree-days are clipped.
        Using 0 will return only positive DD values. The default is 0.

    Returns
    -------
    pd.Series :
        The timeseries of daily degree-days for the period covered by ``temperature``.


    .. warning::

        Days without data receive ``nan`` values, while those with partial data are
        processed as regular ones.


    .. note:: To developers

        The min_count parameter of the sum method is used to return nan when
        one of the values (usually both) is nan. This is the desired behavior,
        as it avoids returning the value of `reference` when the data is not available.

    Examples
    --------
    >>> temp = pd.Series(np.random.randn(100), index=pd.date_range("2020-01-01",
    ... periods=100, freq='h'))
    >>> dd_min_max(temp, 17).head()
    2020-01-01    16.766840
    2020-01-02    17.219413
    2020-01-03    17.168598
    2020-01-04    16.263783
    2020-01-05    17.786569
    Freq: D, Name: heating_degree_days, dtype: float64

    """
    _assert_dd_type(dd_type)
    min_max = temperature.resample("D").apply(["min", "max"])
    if isinstance(min_max, pd.Series):
        min_max = min_max.to_frame().T
    degree_days = reference - min_max.sum(axis=1, min_count=2) / 2
    if dd_type == "cooling":
        degree_days = -degree_days
    degree_days.name = (
        eat.keywords.heating_dd_f if dd_type == "heating" else eat.keywords.cooling_dd_f
    )
    return degree_days.clip(lower=clip_tshd)


def dd_pro(
    temperature: pd.Series,
    reference: float,
    dd_type: literal_valid_dd_types,
    clip_tshd: float = 0,
) -> pd.Series:
    r"""Return daily degree-days with pro method.

    .. math::
        DD = (T_{max} - T_{ref}) * \\
            (0.08 + 0.42 * \\
            (T_{max} - T_{ref}) / (T_{max} - T_{min}))

    Use in France to estimate the degree days during the intermediate seasons,
    when the reference temperature is between the min and max temperature of the day.

    Reference: `meteo France <http://climatheque.meteo.fr/Docs/DJC-methode.pdf>`_

    A small modification is made to the original formula to use the true average
    temperature instead of the min and max temperature mean.
    """
    _assert_dd_type(dd_type)
    min_max_mean = temperature.resample("D").apply(["min", "max", "mean"])
    if dd_type == "cooling":
        # Invert all the signs to compute cooling degree days
        min_max_mean = -min_max_mean
        min_max_mean["min"], min_max_mean["max"] = (
            min_max_mean["max"],
            min_max_mean["min"],
        )
        reference = -reference
        clip_tshd = -clip_tshd
    mask_tmin_over_tref = min_max_mean["min"] > reference
    mask_between_tmin_tmax = (min_max_mean["min"] <= reference) & (
        min_max_mean["max"] >= reference
    )
    degree_days = reference - min_max_mean["mean"]
    degree_days[mask_tmin_over_tref] = clip_tshd
    degree_days[mask_between_tmin_tmax] = (reference - min_max_mean["min"]) * (
        0.08
        + 0.42
        * (reference - min_max_mean["min"])
        / (min_max_mean["max"] - min_max_mean["min"])
    )
    degree_days.name = (
        eat.keywords.heating_dd_f if dd_type == "heating" else eat.keywords.cooling_dd_f
    )
    return degree_days


def dd_mean(
    temperature: pd.Series,
    reference: float,
    dd_type: literal_valid_dd_types,
    clip_tshd: float = 0,
) -> pd.Series:
    """Return daily degree-days with mean method.

    This uses a better estimate of the daily average temperature than the
    min-max method of :py:func:`dd_min_max`.

    .. math::
        DD = (T_{mean} - T_{ref})^+

    Parameters
    ----------
    temperature : pd.Series
        The timeseries of temperature measures from which DD data has to be inferred.
    reference : float
        The reference temperature for  degree-days computation.
    dd_type : {'heating', 'cooling'}
        The type of degree-days to compute.
    clip_tshd : float or None, optional
        A threshold under which the computed degree-days are clipped.
        Using 0 will return only positive DD values. The default is 0.

    Returns
    -------
    pd.Series :
        The timeseries of daily degree-days for the period covered by ``temperature``.


    .. warning::

        Days without data receive ``nan`` values, while those with partial data are
        processed as regular ones.

    Examples
    --------
    >>> temp = pd.Series(np.random.randn(100), index=pd.date_range("2020-01-01",
    ... periods=100, freq='h'))
    >>> dd_mean(temp, 17).head()
    2020-01-01    16.688565
    2020-01-02    17.205562
    2020-01-03    16.898472
    2020-01-04    16.794069
    2020-01-05    16.808996
    Freq: D, Name: heating_degree_days, dtype: float64

    """
    _assert_dd_type(dd_type)
    degree_days = reference - temperature.resample("D").mean()
    if dd_type == "cooling":
        degree_days = -degree_days
    degree_days.name = (
        eat.keywords.heating_dd_f if dd_type == "heating" else eat.keywords.cooling_dd_f
    )
    return degree_days.clip(lower=clip_tshd)


def dd_integral(
    temperature: pd.Series,
    reference: float,
    dd_type: literal_valid_dd_types,
    clip_tshd: float = 0,
    intraday_clip_tshd: float = 0,
) -> pd.Series:
    r"""Return daily degree-days with integral method.

    This is the most accurate method for computing degree-days, as it takes into
    account the temperature variations during the day.

    .. math::
        DD = \\frac{\\sum_{i=1}^{N} (T_{ref} - T_i)^+  \\
            \\cdot \\Delta t_i}{\\sum_{i=1}^{N} \\Delta t_i}

    Parameters
    ----------
    temperature : pd.Series
        The timeseries of temperature measures from which DD data has to be inferred.
    reference : float
        The reference temperature for  degree-days computation.
    dd_type : {'heating', 'cooling'}
        The type of degree-days to compute.
    clip_tshd : float or None, optional
        A threshold under which the computed degree-days are clipped.
        Using 0 will return only positive DD values. The default is 0.
    intraday_clip_tshd : float or None, optional
        A threshold under which the instantaneous temperature differences used
        in degree-days computation are clipped.
        If None, no clipping is done, hence resulting in the same results as
        :py:func:`dd_mean`. The default is ``0``.
        Using ``0`` will assign no heating-need (resp. cooling) to moments
        when the reference temperature is lower than the outside one.

    Returns
    -------
    pd.Series :
        The timeseries of daily degree-days for the period covered by ``temperature``.


    .. warning::

        Days without data receive ``nan`` values, while those with partial data are
        processed as regular ones.

    Notes
    -----
    About the parameter ``intraday_clip_tshd``:

    If the temperature is always below or above the reference temperature, the clipping
    has no effect.

    If the temperature is sometimes below and sometimes above the reference temperature,
    The question is : does the cold temperature compensate the warm temperature?
    Meaning: would the user heat (resp. cool) the building if the temperature in the
    morning, even though it is warm in the afternoon? If the answer is yes, then the
    clipping should be set to 0. If the answer is no, then the clipping should be set
    to ``None``.

    Examples
    --------
    >>> temp = pd.Series(np.random.randn(100), index=pd.date_range("2020-01-01",
    ... periods=100, freq='h'))
    >>> dd_integral(temp, 17).head()
    2020-01-01    16.666671
    2020-01-02    17.109244
    2020-01-03    17.087415
    2020-01-04    16.976660
    2020-01-05    16.337652
    Freq: D, Name: heating_degree_days, dtype: float64

    """
    _assert_dd_type(dd_type)
    timesteps = eat.timeseries.extract_features.timestep_durations(temperature)
    sign = 1 if dd_type == "heating" else -1
    degree_days = (
        (sign * (reference - temperature)).clip(lower=intraday_clip_tshd) * timesteps
    ).resample("D").sum() / timesteps.resample("D").sum()
    degree_days.name = (
        eat.keywords.heating_dd_f if dd_type == "heating" else eat.keywords.cooling_dd_f
    )
    return degree_days.clip(lower=clip_tshd)


def dd_compute(
    temperature: pd.Series,
    reference: float,
    dd_type: literal_valid_dd_types,
    clip_tshd: float = 0,
    method: literal_computation_dd_types = "integral",
    **kwargs,
) -> pd.Series:
    """Return daily degree-days with the specified method.

    Parameters
    ----------
    temperature : pd.Series
        The
    reference : float
        The reference temperature for  degree-days computation.
    dd_type : {'heating', 'cooling'}
        The type of degree-days to compute.
    clip_tshd : float or None, optional
        A threshold under which the computed degree-days are clipped.
        Using 0 will return only positive DD values. The default is 0.
    method : {'min_max', 'mean', 'integral'}, optional
        The method to use for computing the degree-days. The default is 'integral'.
        - 'min_max' : Min-max method :py:func:`dd_min_max`
        - 'mean' : Mean method :py:func:`dd_mean`
        - 'integral' : Integral method :py:func:`dd_integral`
        - 'pro' : Pro method :py:func:`dd_pro`
    kwargs : mapping, optional
        A dictionary of keyword arguments passed to the degree-days computation
        methods, such as :py:func:`dd_integral`.

    Returns
    -------
    pd.Series :
        The timeseries of daily degree-days for the period covered by ``temperature``.

    Raises
    ------
    EATInvalidDegreeDaysMethodError
        If the method is not recognized.

    See Also
    --------
    - :py:func:`dd_min_max` : Min-max method
    - :py:func:`dd_mean` : Mean method
    - :py:func:`dd_integral` : Integral method
    - :py:func:`dd_pro` : Pro method

    Examples
    --------
    >>> temp = pd.Series(np.random.randn(100), index=pd.date_range("2020-01-01",
    ... periods=100, freq='h'))
    >>> dd_compute(temp, 17, method="integral").head()
    2020-01-01    16.968044
    2020-01-02    17.147927
    2020-01-03    16.837810
    2020-01-04    16.877662
    2020-01-05    17.453723
    Freq: D, Name: heating_degree_days, dtype: float64

    """
    methods: dict[str, DegreeDaysFunction] = {
        "min_max": dd_min_max,
        "mean": dd_mean,
        "integral": dd_integral,
        "pro": dd_pro,
    }
    try:
        func: DegreeDaysFunction = methods[method]
    except KeyError:
        raise EATInvalidDegreeDaysMethodError(method, list(methods.keys())) from None
    return func(temperature, reference, dd_type, clip_tshd, **kwargs)


def dd_calc_method(
    func: Callable,
) -> Literal["min_max", "mean", "integral", "pro", "unknown"]:
    """Return the name of the method used for computing degree-days.

    Parameters
    ----------
    func : function
        The function to check.

    Returns
    -------
    str :
        The name of the method used for computing degree-days.

    """
    func_to_method = {
        dd_min_max: "min_max",
        dd_mean: "mean",
        dd_integral: "integral",
        dd_pro: "pro",
    }
    if func in func_to_method:
        return cast(Literal["min_max", "mean", "integral", "pro"], func_to_method[func])
    return "unknown"


def _assert_dd_type(
    dd_type: str,
) -> None:
    """Check that the degree-days type is valid.

    Parameters
    ----------
    dd_type : str
        the type of degree-days to compute.

    Raises
    ------
    EATInvalidDegreeDaysError
        If the type is not recognized, i.e. not 'heating' or 'cooling'.

    """
    valid_dd_types = ["heating", "cooling"]
    if dd_type not in valid_dd_types:
        raise EATInvalidDegreeDaysError(
            dd_type,
            valid_dd_types,
        )



import numpy as np
import pandas as pd

from ..power.overconsumption.find import from_power_threshold


def example_power():
    """Return an example power series, threshold and overshoot overconsumption.

    The power is sampled with 30min period such that any error on the bounds by
    +/- 1 step is pretty obvious in terms of energy.
    The power is piecewise-constant, equal to 1W everywhere except :

    - during 1h from 10:00 to 11:00 when it is 11W
    - during 1h30min from 14:00 to 15:30 when it is 21W.

    """
    times = pd.date_range(
        start=pd.Timestamp("2023-03-26", tz="Europe/Paris"),  # DST
        end=pd.Timestamp("2023-03-27", tz="Europe/Paris"),
        inclusive="left",
        freq="30min",
    )
    pow = pd.Series(np.ones_like(times), index=times)
    s1 = pd.Timestamp("2023-03-26 10:00", tz="Europe/Paris")
    e1 = pd.Timestamp("2023-03-26 10:30", tz="Europe/Paris")
    s2 = pd.Timestamp("2023-03-26 14:00", tz="Europe/Paris")
    e2 = pd.Timestamp("2023-03-26 15:00", tz="Europe/Paris")
    threshold = pow.copy()
    pow.loc[s1:e1] = 11  # inclusive loc
    pow.loc[s2:e2] = 21  # inclusive loc
    intervals = pd.DataFrame.from_dict(
        {
            "start": [s1, s2],
            "end": [e1, e2],
            "duration": [3600, 5400],
            "energy": [3600 * 10, 5400 * 20],
        },
    )
    intervals["end"] += pd.Timedelta("30min")  # doom pandas inclusive loc
    return pow, threshold, intervals


def test_from_power_threshold_constant():
    """Check the function with constant threshold as float and series"""
    power, threshold, intervals = example_power()
    # constant as series
    found_intervals = from_power_threshold(
        power,
        overshoot_tshd=threshold,
    )
    pd.testing.assert_frame_equal(found_intervals, intervals, check_dtype=False)
    # constant as float
    found_intervals = from_power_threshold(
        power,
        overshoot_tshd=1.0,
    )
    pd.testing.assert_frame_equal(found_intervals, intervals, check_dtype=False)


def test_from_power_threshold_over_power():
    """Check the function with threshold over max power"""
    power, _, _ = example_power()
    assert from_power_threshold(power, 100.0).empty
    assert from_power_threshold(power, power.max()).empty


def test_from_power_threshold_under_power():
    """Check the function with threshold under min power.

    Double limit-case :

    - DST : the day is shorter than 24h
    - Whole day : the end of the overconsumption is the last instant in the
      series, as the moment after is not known. (should be the end of the timestep).
    """
    power, _, intervals = example_power()
    found_intervals = from_power_threshold(power, 0.0)
    intervals_expect = pd.DataFrame.from_dict(
        {
            "start": [power.index[0]],
            "end": [power.index[-1]],
            "duration": 1800 * 45,
            "energy": 1800 * 45 + intervals["energy"].sum(),
        },
    )
    pd.testing.assert_frame_equal(found_intervals, intervals_expect, check_dtype=False)


def test_from_power_threshold_series():
    """Check that the function works correctly with a variable threshold"""
    power, threshold, intervals = example_power()
    threshold.loc[: pd.Timestamp("2023-03-26 12:00", tz="Europe/Paris"),] = 11
    found_intervals = from_power_threshold(
        power,
        overshoot_tshd=threshold,
    )
    pd.testing.assert_frame_equal(
        found_intervals, intervals[1:].reset_index(drop=True), check_dtype=False,
    )


def test_from_power_threshold_custom_ref():
    """Check that the custom energy reference is correctly accounted for"""
    power, threshold, intervals = example_power()
    ref = threshold.copy()
    threshold.loc[:] = 11
    found_intervals = from_power_threshold(
        power,
        overshoot_tshd=threshold,
        reference_energy_tshd=ref,
    )
    pd.testing.assert_frame_equal(
        found_intervals, intervals[1:].reset_index(drop=True), check_dtype=False,
    )



"""Provide constants used throughout the package."""

MINUTE = 60
DAY = 24 * 60 * MINUTE

#: The number of Joules in a Watt-hour.
wh_to_j = 3600
#: The number of Joules in a kiloWatt-hour.
kwh_to_j = 1000 * wh_to_j



"""Package to resample timeseries."""

from ._facade import (
    to_freq,
    trim_out_of_bounds,
)
from .conservative import (
    flow_rate_conservative,
    flow_rate_to_freq,
    volume_conservative,
    volume_to_freq,
)
from .index_transformation import (
    estimate_timestep,
    fill_data_holes,
    fill_missing_entries,
    index_to_freq,
    tz_convert_or_localize,
)
from .interpolate import (
    piecewise_affine,
    piecewise_constant,
)



"""Compute hybrid profiles based on deviations from the mean profile.

This module provides tools to calculate threshold profiles driven by a combination
of standard deviation and relative thresholds derived from historical data.
The thresholds are designed to adapt to variability in the data, balancing stability
with tolerance for variability in consumption patterns.
"""

import pandas as pd

from energy_analysis_toolbox.timeseries.profiles.mean_profile import MeanProfile


class HybridThreshold(MeanProfile):
    """Returns the deviation from mean profile (hybrid std and relative thresholds)."""

    def __init__(
        self,
        offset_std: float = 3,
        offset_relative: float = 0.5,
        **kwargs,
    ) -> None:
        """Initialize HybridThreshold.

        Parameters
        ----------
        period : str, optional
            A pandas period string which specifies the kind of period on which
            the profile is realized.
        offset_std : float, optional
            Number of standard deviations VS the computed reference to obtain
            the threshold profile. Default is 3 (profile is 3 standard deviations
            from reference)
        offset_relative : float, optional
            Relative difference VS the computed reference to obtain
            the threshold profile. Default is 0.5 (profile is 150%
            of reference)

        """
        super().__init__(**kwargs)
        self.offset_std = offset_std
        self.offset_rel = offset_relative

    def compute(
        self,
        history: pd.Series,
        time: pd.Timestamp,
        **kwargs,
    ) -> pd.Series:
        """Return a threshold profile.

        The threshold profile is obtained using a user-defined relative variation
        from the mean profile built from history.

        Parameters
        ----------
        history : pd.Series
            Consumption history used to computed the reference
            profile.
        time : pd.Timestamp
            The time at which the profile is of interest. Only
            the information about the date is used in the passed
            timestamp.

        Returns
        -------
        profile : pd.Series
            Profile threshold with same resolution as the history data

        Notes
        -----
        The profile threshold is obtained as an hybrid between the std and
        arbitrary tshd profiles :

            - an average profile is computed from the rolling max of the history
              on a window centered on each slot with size ``window``.
            - a std profile is computed from the history (without rolling max)
            - the returned profile is obtained as the max between :

                * the profile based on ``tshd_std`` standard deviation over the
                  mean profile (the std is computed from the data without rolling
                  aggregation),
                * the profile based on ``(1 + tshd_rel)`` times the mean profile.

        The idea behind this profile is to base the threshold on the data variability,
        but to still have a certain tolerance when the consumption is very stable in
        the (short) history.

        The rolling max also artificially increases the mean value around slots where
        the consumption is high in order to account for the correlation between the
        consumption on consecutive slots : a certain "horizontal" dispersion of the
        consumption is relatively normal.


        .. note:

            In ``self.is_max == False``, replace "max" by "min" in the text above.

        """
        # rel and ref on smoothed data
        smooth_mean_profile = super().compute(history, time, **kwargs)
        rel_profile = smooth_mean_profile * self.offset_rel
        # std on unsmoothed data
        std_profile = self.group(history).std() * self.offset_std
        std_profile.index = rel_profile.index
        profile_deviations = pd.DataFrame.from_dict(
            {"std": std_profile, "tshd": rel_profile},
        )
        if self.is_max:
            var_profile = profile_deviations.max(axis=1)
        else:
            var_profile = profile_deviations.min(axis=1)
        return smooth_mean_profile + var_profile



"""Defines a base class to compute average load profiles from history."""

import pandas as pd


class MeanProfile:
    """A class which computes a simple mean profile."""

    def __init__(
        self,
        period: str = "D",
        *,
        is_max: bool = True,
        window: int = 1,
    ) -> None:
        """Initialize a MeanProfile instance.

        Parameters
        ----------
        period : str, optional
            A pandas period string which specifies the kind of period on which the
            profile is realized.
        is_max : bool, optional
            The type of threshold defined by this profile. If True, defines a
            threshold which defines a maximum value which should not be
            crossed, while False defines a minimum value.
        window : int, optional
            Width (# of slots) of the rolling max window used to transform
            the history used to deduce the mean profile (but not the std).
            Default is 1 slots (meaning no transformation).

        """
        self.period = period
        self.is_max = is_max
        self.window = window

    def group(
        self,
        history: pd.Series | pd.DataFrame,
    ) -> pd.core.groupby.GroupBy:
        """Group history data by chunks of ``self.period``.

        Parameters
        ----------
        history : pd.Series or pd.DataFrame

        Returns
        -------
        pd.core.groupby.GroupBy

        """
        return history.groupby(history.index - history.index.floor(self.period))

    def compute(
        self,
        history: pd.Series,
        time: pd.Timestamp,
    ) -> pd.Series:
        """Return an averaged profile.

        Parameters
        ----------
        history : pd.Series
            Consumption history used to computed the reference
            profile. The history should be sampled homogeneously during the
            aggregated periods.
        time : pd.Timestamp
            The time at which the profile is of interest. Only
            the information about the date floored to ``self.period`` is
            used in the passed timestamp.

        Returns
        -------
        profile : pd.Series
            Profile threshold with same resolution as the history data.

        Notes
        -----
        The profile threshold is obtained as the mean profile obtained
        from history.

        """
        if self.window == 1:
            pass
        elif self.is_max:
            history = history.rolling(self.window, center=True).max()
        else:
            history = history.rolling(self.window, center=True).min()
        mean_profile_compare = self.group(history).mean(numeric_only=True)
        mean_profile_compare.index += time
        return mean_profile_compare



"""Defines basic custom errors."""


class EATError(Exception):
    """The base class for all except in |eat|.

    All exceptions of |eat| library should inherit this class, such that exceptions
    specific to this lib can easily be caught/identified.
    """


class EATEmptyDataError(EATError):
    """An empty data container is passed, but this case cannot be managed."""



"""Contains the definition of "keywords"/ "magic values" used in library's public API.

All variables in this module are basic python objects such as int, str, list etc.

Note about the naming conventions in this module
------------------------------------------------
In order to have readable yet as light as possible names, the following
conventions are used in this module:

1. Variables names are written in snake-case
2. Names suffixed with "_f" are reserved for variables containing field
   names used to identify values in tables such as `pandas` dataframes
3. Names suffixed with "_l" are reserved for sequences (lists) of keywords

.. important::

    DO NOT use these keywords by typing directly their VALUES. The preferred
    use is to pass the variables defined in this module, so that:

    - your code will not be broken in case the content of the variable
      (the magic values) are modified
    - any misspelling will be caught by Python interpreter and lead to
      a ``NameError``
    - you can benefit from the help of autocompletion tools provide
      by your IDE if you have one

Variables defined by the module
-------------------------------
The following reproduces the variables declarations from the module.
"""

#: The name of a field containing the beginning of a time-interval. Unless stated
#: otherwise, the first instant to be included in the interval, as a timestamp.
start_f = "start"
#: The name of a field containing the end of a time-interval. Unless stated
#: otherwise, the first instant to be excluded in the interval, as a timestamp.
end_f = "end"
#: The name of the time-index in timeseries data.
time_f = "timestamp"
#: The name of the field containing the heating degree-days.
heating_dd_f = "heating_degree_days"
#: The name of the field containing the cooling degree-days.
cooling_dd_f = "cooling_degree_days"
#: The name of the field containing the base consumption,
#: i.e. the consumption that is not affected by the weather.
base_f = "base"



"""Define Mixins to manage localized timeseries."""

import pandas as pd

from .mean_profile import MeanProfile
from .rolling_profile import (
    RollingProfile,
    RollingQuantileProfile,
)


class LocalizedProfileMixin:
    """A profile version where time-zoned data is managed, including DSTs.

    The compute method of the profile class with which this mixin is combined
    is overloaded in such a way that passing time-localized data is possible.
    On and across DSTs, the data remains aligned "on the clock" (VS on the sun).
    """

    def compute(
        self,
        history: pd.DataFrame,
        time: pd.Timestamp,
        **kwargs,
    ) -> pd.DataFrame:
        """Compute the profile at ``time`` from the ``history``.

        The computation is managed as follows :

        - just fall back to the parent `compute` method in case the data is time-naive.
        - Else, unlocalize the history and target date before passing them to the
          parent class' `compute` such that the data remains aligned based on "local
          time" for each day.
        - Relocalize the resulting profile to the history timezone. This part takes
          care of DST days :

          * In case the target date is summer DST (23h long day) the extra profile
            hours is just dumped.
          * In case the target date is winter DST(25h long day), the data for the 2am
            to 3am period is associated with the first occurrence of this time-of-day
            and a gap is let on the second.


        .. warning::

          If winter DST is included in the history, the timestamps from 2am to
          3am on this day appears "twice" in the time-naive history passed to the
          aggregation. Therefore, the aggregation should be robust to this situation.


        .. seealso::

            :py:meth:`.MeanProfile.compute`

        """
        source_tz = history.index.tz
        if source_tz is None:
            profile_ref = super().compute(history, time, **kwargs)
        else:
            profile_ref = (
                super()
                .compute(history.tz_localize(None), time.tz_localize(None), **kwargs)
                .tz_localize(source_tz, ambiguous=True, nonexistent="NaT")
            )
        return profile_ref


class LocalizedMeanProfile(
    LocalizedProfileMixin,
    MeanProfile,
):
    """Compute the mean profile with timezone data.

    see :py:class:`.MeanProfile`
    """


class LocalizedRollingProfile(
    LocalizedProfileMixin,
    RollingProfile,
):
    """Compute a rolling profile with timezone data.

    see :py:class:`.RollingProfile`
    """


class LocalizedRollingQuantileProfile(
    LocalizedProfileMixin,
    RollingQuantileProfile,
):
    """Compute a rolling quantile with timezone data.

    see :py:class:`.RollingQuantProfile`
    """



"""Tests for :py:mod:`energy_analysis_toolbox.timeseries.resample._facade` module."""

import pandas as pd

from ..timeseries.resample._facade import to_freq

example_series = pd.Series(
    [0.0, 1.0],
    index=pd.date_range("2020-01-01 01:00", periods=2, freq="1h"),
    name="example_series",
)


def test_to_freq_piecewise_affine():
    """The function checks the default interpolation strategy."""
    # oversampling
    pd.testing.assert_series_equal(
        to_freq(example_series, "30min"),
        pd.Series(
            [0, 0.5, 1, 1],
            index=pd.date_range("2020-01-01 01:00", periods=4, freq="30min"),
            name="example_series",
        ),
    )
    # subsampling
    pd.testing.assert_series_equal(
        to_freq(example_series, "2h"),
        pd.Series(
            [0.0],
            index=pd.date_range("2020-01-01 01:00", periods=1, freq="2h"),
            name="example_series",
        ),
    )
    pd.testing.assert_series_equal(
        to_freq(example_series, "2h", origin="floor"),
        pd.Series(
            [0.0, 1.0],
            index=pd.date_range("2020-01-01 00:00", periods=2, freq="2h"),
            name="example_series",
        ),
    )
    pd.testing.assert_series_equal(
        to_freq(example_series, "3h", origin="ceil"),
        pd.Series(
            [1.0],
            index=pd.date_range("2020-01-01 03:00", periods=1, freq="3h"),
            name="example_series",
        ),
    )
    assert to_freq(example_series.iloc[:0], "1h").empty


def test_to_freq_piecewise_constant():
    """The function checks the default interpolation strategy."""
    # oversampling
    pd.testing.assert_series_equal(
        to_freq(example_series, "30min", method="piecewise_constant"),
        pd.Series(
            [0.0, 0.0, 1, 1],
            index=pd.date_range("2020-01-01 01:00", periods=4, freq="30min"),
            name="example_series",
        ),
    )
    # subsampling
    pd.testing.assert_series_equal(
        to_freq(example_series, "2h", method="piecewise_constant"),
        pd.Series(
            [0.0],
            index=pd.date_range("2020-01-01 01:00", periods=1, freq="2h"),
            name="example_series",
        ),
    )
    pd.testing.assert_series_equal(
        to_freq(example_series, "2h", origin="floor", method="piecewise_constant"),
        pd.Series(
            [
                0.0,
                1,
            ],
            index=pd.date_range("2020-01-01 00:00", periods=2, freq="2h"),
            name="example_series",
        ),
    )
    pd.testing.assert_series_equal(
        to_freq(example_series, "3h", origin="ceil", method="piecewise_constant"),
        pd.Series(
            [1.0],
            index=pd.date_range("2020-01-01 03:00", periods=1, freq="3h"),
            name="example_series",
        ),
    )
    assert to_freq(example_series.iloc[:0], "1h").empty


def test_to_freq_volume_conservative():
    """The function just checks that the right method is called as it is just a
    wrapper.

    """
    # oversampling
    pd.testing.assert_series_equal(
        to_freq(
            example_series,
            "30min",
            method="volume_conservative",
            last_step_duration=3600 * 2,
        ),
        pd.Series(
            [0.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            index=pd.date_range("2020-01-01 01:00", periods=6, freq="30min"),
            name="example_series",
        ),
    )


def test_to_freq_flow_rate_conservative():
    """The function just checks that the right method is called as it is just a
    wrapper.

    """
    # oversampling
    pd.testing.assert_series_equal(
        to_freq(
            example_series,
            "30min",
            method="flow_rate_conservative",
            last_step_duration=3600 * 2,
            origin=pd.Timestamp("2020-01-01 00:00"),
        ),
        pd.Series(
            [0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0],
            index=pd.date_range("2020-01-01 00:00", periods=8, freq="30min"),
            name="example_series",
        ),
    )


def test_to_freq_custom_method():
    """Test the function when passing a custom resamplig function."""

    def custom(timeseries, *args, **kwargs):
        return timeseries.resample(kwargs["freq"], origin="start").mean()

    pd.testing.assert_series_equal(
        to_freq(example_series, "2h", method=custom),
        pd.Series(
            [
                0.5,
            ],
            index=pd.date_range("2020-01-01 01:00", periods=1, freq="2h"),
            name="example_series",
        ),
    )



"""Applies elementary operations on power timeseries."""

import pandas as pd

from energy_analysis_toolbox.timeseries.extract_features import timestep_durations


def integrate_over(
    intervals: pd.DataFrame,
    power_series: pd.Series,
) -> pd.Series:
    """Return the energy integrated over slices of the power series.

    .. warning::

         This function only works on slices of the power series.
         It cannot be used to integrate the power with smaller
         granularity.


    Parameters
    ----------
    intervals : pd.DataFrame
        A table of overconsumption defined with 'start' and 'end'
        columns containing timestamps. **These timestamps are used as slice**
        **bounds in the series** with included start and excluded end.
    power_series : pd.Series
        A timeseries of power in (W). The series must have at least
        two elements so that the timestep(s) can be deduced.

    Returns
    -------
    energies : pd.Series
        A series of energy in (J) with same index as ``overconsumption``.


    .. note::

        Proper integration with a resolution independent from the timestep can
        be achieved by transforming the power to energy using :py:func:`to_energy`
        and then using the |volume_conservative| resampling.


    Examples
    --------
    >>> power = _constant_power(); power
    2023-10-29 00:00:00+02:00    1
    2023-10-29 01:00:00+02:00    1
    ...
    2023-10-30 22:00:00+01:00    1
    2023-10-30 23:00:00+01:00    1
    Freq: H, dtype: object
    >>> overconsumption = _intervals()
    >>> overconsumption
                           start                        end
    A  2023-10-29 00:00:00+02:00  2023-10-29 01:00:00+02:00
    B  2023-10-30 02:00:00+02:00  2023-10-30 02:00:00+01:00
    >>> eat.power.integrate_over(overconsumption, power)
    A    3600.0
    B    3600.0
    dtype: float64

    Remind that the interval bounds are used to slice the series (with excluded
    end). Accordingly, an interval is completely accounted for in the integration
    as soon as it falls with the slice :
    >>> overconsumption['end'] += pd.Timedelta('10min')
    >>> eat.power.integrate_over(overconsumption, power)
    A    7200.0
    B    7200.0
    dtype: float64


    """
    timesteps = timestep_durations(power_series)

    def integrate(row: pd.Series) -> float:
        """Integrate on a slice with excluded end."""
        i_start = timesteps.index.get_slice_bound(row["start"], side="left")
        i_end = timesteps.index.get_slice_bound(row["end"], side="left")
        return (timesteps.iloc[i_start:i_end] * power_series.iloc[i_start:i_end]).sum()

    return intervals.apply(integrate, axis=1)


def to_energy(
    power_series: pd.Series,
) -> pd.Series:
    """Return a series of energy per timestep.

    Parameters
    ----------
    power_series : pd.Series
        Timeseries of (avg) power per ``[ti, ti+1[`` timestep in (W).
        The series must have at least two elements so that the timestep(s)
        can be deduced. The last timestep is assumed to be the same as the
        former last one.

    Returns
    -------
    pd.Series :
        Timeseries of energy per timestep in (J) with same index
        as input series.

    """
    timesteps = timestep_durations(power_series)
    return timesteps * power_series



"""Module converting tabular representations of interval data into timeseries."""

import numpy as np
import pandas as pd

from energy_analysis_toolbox import keywords as eatk


def flatten_and_fill(
    data: pd.DataFrame,
    fill_values: dict | None = None,
    start_f: str = eatk.start_f,
    end_f: str = eatk.end_f,
    time_f: str = eatk.time_f,
) -> pd.DataFrame:
    """Return data as a table of timeseries.

    Parameters
    ----------
    data : pd.DataFrame
        A table of overconsumption which are each defined by a ``start_f``, used as
        index, and a ``end_f``.
    fill_values : dict or None
        A dict where keys are names of columns of ``data`` and values are
        used to fill the duration between the overconsumption for each of the column
        in the dict. Missing column receive ``np.nan``.
        Default is |None| which is treated as an empty dict.
    start_f : str, default |eatk.start_f|
        The name of the column defining interval starts.
    end_f : str, default |eatk.end_f|
        The name of the column defining interval ends.
    time_f : str, default |eatk.end_f|
        The name of the index of the returned Dataframe.

    Returns
    -------
    pd.DataFrame :
        A table describing the input data as a timeseries.
        The dataframe is indexed with the times of the interval starts and ends,
        such that the rows indexed with an interval end in the dataframe contains
        only filler values.


    .. note::

        An empty table is considered to be flattened as an empty table.


    .. warning::

        The function assumes that the overconsumption are correctly defined meaning
        that :

        - no interval overlap one each other,
        - overconsumption have no common boundary.

        The function does not check these assumptions.

    Example
    -------
    Consider e.g. the following dataframe (from the function tests).

    >>> table
                                        end col_str  col_float
    timestamp
    2022-12-05 18:00:00 2022-12-05 18:05:00    toto          0
    2022-12-05 18:10:00 2022-12-05 18:15:00    toto          2
    2022-12-05 18:20:00 2022-12-05 18:25:00    toto          4

    It can be flattened as follows :

    >>> flatten_and_fill(table)
                        col_str  col_float
    timestamp
    2022-12-05 18:00:00    toto        0.0
    2022-12-05 18:05:00     NaN        NaN
    2022-12-05 18:10:00    toto        2.0
    2022-12-05 18:15:00     NaN        NaN
    2022-12-05 18:20:00    toto        4.0
    2022-12-05 18:25:00     NaN        NaN

    The ``nan`` values can be filled with fixed values for each column during
    the flattening process :

    >>> flatten_and_fill(table, fill_values={'col_str': 'tata', 'col_float': 0})
                        col_str  col_float
    timestamp
    2022-12-05 18:00:00    toto          0
    2022-12-05 18:05:00    tata          0
    2022-12-05 18:10:00    toto          2
    2022-12-05 18:15:00    tata          0
    2022-12-05 18:20:00    toto          4
    2022-12-05 18:25:00    tata          0

    """
    if data.empty:
        return data
    fillers = data.set_index(end_f).drop(columns=[start_f])
    fill_inputs = {col: np.nan for col in fillers.columns}
    if fill_values is not None:
        fill_inputs.update(fill_values)
        for col, value in fill_inputs.items():
            new_col = pd.Series(
                data=value,
                index=fillers.index,
                name=col,
                dtype=fillers.dtypes[col],
            )
            fillers.loc[:, col] = new_col
    else:
        for col in fillers.columns:
            fillers.loc[:, col] = pd.Series(
                data=np.nan,
                index=fillers.index,
                name=col,
            ).astype(fillers.dtypes[col])
    timeseries_table = pd.concat(
        [data.set_index(start_f).drop(columns=[end_f]), fillers],
    ).sort_index()
    timeseries_table.index.name = time_f
    return timeseries_table
